import pickle
import re
from io import BytesIO
from typing import IO, Iterator

from bson.binary import Binary

from minimalkv._key_value_store import KeyValueStore


class MongoStore(KeyValueStore):
    """Uses a MongoDB collection as the backend, using pickle as a serializer.

    Parameters
    ----------
    db :
        An authenticated pymongo database.
    collection : str
        A MongoDB collection name.

    """

    def __init__(self, db, collection):

        self.db = db
        self.collection = collection

    def _has_key(self, key: str) -> bool:
        return self.db[self.collection].count_documents({"_id": key}) > 0

    def _delete(self, key: str) -> str:
        return self.db[self.collection].delete_one({"_id": key})

    def _get(self, key: str) -> bytes:
        try:
            item = next(self.db[self.collection].find({"_id": key}))
            return pickle.loads(item["v"])
        except StopIteration:
            raise KeyError(key)

    def _open(self, key: str) -> IO:
        return BytesIO(self._get(key))

    def _put(self, key: str, value: bytes) -> str:
        self.db[self.collection].update_one(
            {"_id": key}, {"$set": {"v": Binary(pickle.dumps(value))}}, upsert=True
        )
        return key

    def _put_file(self, key: str, file: IO) -> str:
        return self._put(key, file.read())

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        for item in self.db[self.collection].find(
            {"_id": {"$regex": "^" + re.escape(prefix)}}
        ):
            yield item["_id"]



import os

import pytest

boto3 = pytest.importorskip("boto3")
from io import BytesIO

from basic_store import BasicStore
from bucket_manager import boto3_bucket, boto_credentials
from conftest import ExtendedKeyspaceTests
from url_store import UrlStore

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.net.boto3store import Boto3Store


@pytest.fixture(
    params=boto_credentials, ids=[c["access_key"] for c in boto_credentials]
)
def credentials(request):
    return request.param


@pytest.fixture
def bucket(credentials):
    with boto3_bucket(**credentials) as bucket:
        yield bucket


class TestBoto3Storage(BasicStore, UrlStore):
    @pytest.fixture(params=[True, False])
    def reduced_redundancy(self, request):
        return request.param

    @pytest.fixture
    def storage_class(self, reduced_redundancy):
        return "REDUCED_REDUNDANCY" if reduced_redundancy else None

    @pytest.fixture(params=["", "/test-prefix"])
    def prefix(self, request):
        return request.param

    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        return Boto3Store(bucket, prefix, reduced_redundancy=reduced_redundancy)

    # Disable max key length test as it leads to problems with minio
    test_max_key_length = None

    def test_get_filename_nonexistant(self, store, key, tmp_path):
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_key_error_on_nonexistant_get_filename(self, store, key, tmp_path):
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_storage_class_put(self, store, prefix, key, value, storage_class, bucket):
        store.put(key, value)
        obj = bucket.Object(prefix.lstrip("/") + key)
        assert obj.storage_class == storage_class

    def test_storage_class_putfile(
        self, store, prefix, key, value, storage_class, bucket
    ):
        store.put_file(key, BytesIO(value))
        obj = bucket.Object(prefix.lstrip("/") + key)
        assert obj.storage_class == storage_class


class TestExtendedKeyspaceBoto3Store(TestBoto3Storage, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, Boto3Store):
            pass

        return ExtendedKeyspaceStore(
            bucket, prefix, reduced_redundancy=reduced_redundancy
        )



import inspect
import os
import sys

from sphinx.ext import apidoc

sys.path.append("../")

package = "minimalkv"
html_theme = "alabaster"

__location__ = os.path.join(
    os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore
)

# Generate module references
output_dir = os.path.abspath(os.path.join(__location__, "../docs/_rst"))
module_dir = os.path.abspath(os.path.join(__location__, "..", package))

apidoc_parameters = ["-f", "-e", "-o", output_dir, module_dir]
apidoc.main(apidoc_parameters)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

# General information about the project.
project = "minimalkv"
copyright = "2011-2021, The minimalkv contributors"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "0.14.1"
# The full version, including alpha/beta/rc tags.
release = "0.14.1"

exclude_trees = ["_build"]

pygments_style = "sphinx"

intersphinx_mapping = {
    "http://docs.python.org/": None,
    "http://docs.pythonboto.org/en/latest/": None,
    "http://sendapatch.se/projects/pylibmc/": None,
    "http://www.sqlalchemy.org/docs/": None,
    "http://redis-py.readthedocs.org/en/latest/": None,
}



import pytest

from minimalkv._store_creation import create_store


def test_create_store_azure(mocker):
    # Mock HAzureBlockBlobStore also here, becase otherwise it will try to inherit from
    # the mock object `mock_azure` created below, which will fail.
    mock_hazure = mocker.patch("minimalkv._hstores.HAzureBlockBlobStore")
    mock_azure = mocker.patch("minimalkv.net.azurestore.AzureBlockBlobStore")
    create_store(
        "azure",
        {
            "account_name": "ACCOUNT",
            "account_key": "KEY",
            "container": "cont_name",
            "create_if_missing": True,
        },
    )
    mock_azure.assert_called_once_with(
        checksum=True,
        conn_string="DefaultEndpointsProtocol=https;AccountName=ACCOUNT;AccountKey=KEY",
        container="cont_name",
        create_if_missing=True,
        max_connections=2,
        public=False,
        socket_timeout=(20, 100),
        max_block_size=4194304,
        max_single_put_size=67108864,
    )
    mock_hazure.assert_not_called()


def test_create_store_hazure(mocker):
    mock_hazure = mocker.patch("minimalkv._hstores.HAzureBlockBlobStore")
    create_store(
        "hazure",
        {
            "account_name": "ACCOUNT",
            "account_key": "KEY",
            "container": "cont_name",
            "create_if_missing": True,
        },
    )
    mock_hazure.assert_called_once_with(
        checksum=True,
        conn_string="DefaultEndpointsProtocol=https;AccountName=ACCOUNT;AccountKey=KEY",
        container="cont_name",
        create_if_missing=True,
        max_connections=2,
        public=False,
        socket_timeout=(20, 100),
        max_block_size=4194304,
        max_single_put_size=67108864,
    )


def test_create_store_azure_inconsistent_params():
    with pytest.raises(
        Exception, match="create_if_missing is incompatible with the use of SAS tokens"
    ):
        create_store(
            "hazure",
            {
                "account_name": "ACCOUNT",
                "account_key": "KEY",
                "container": "cont_name",
                "create_if_missing": True,
                "use_sas": True,
            },
        )


def test_create_store_hs3(mocker):
    mock_hs3 = mocker.patch("minimalkv._boto._get_s3bucket")
    create_store(
        "hs3",
        {
            "host": "endpoint:1234",
            "access_key": "access_key",
            "secret_key": "secret_key",
            "bucket": "bucketname",
        },
    )
    mock_hs3.assert_called_once_with(
        host="endpoint:1234",
        access_key="access_key",
        secret_key="secret_key",
        bucket="bucketname",
    )


def test_create_store_s3(mocker):
    mock_s3 = mocker.patch("minimalkv._boto._get_s3bucket")
    create_store(
        "s3",
        {
            "host": "endpoint:1234",
            "access_key": "access_key",
            "secret_key": "secret_key",
            "bucket": "bucketname",
        },
    )
    mock_s3.assert_called_once_with(
        host="endpoint:1234",
        access_key="access_key",
        secret_key="secret_key",
        bucket="bucketname",
    )


def test_create_store_hfs(mocker):
    mock_hfs = mocker.patch("minimalkv._hstores.HFilesystemStore")
    mock_makedirs = mocker.patch("os.makedirs")
    create_store(
        "hfs",
        {"type": "hfs", "path": "this/is/a/relative/path", "create_if_missing": True},
    )
    mock_hfs.assert_called_once_with("this/is/a/relative/path")
    mock_makedirs.assert_called_once_with("this/is/a/relative/path")


def test_create_store_fs(mocker):
    mock_fs = mocker.patch("minimalkv._store_creation.FilesystemStore")
    mock_makedirs = mocker.patch("os.makedirs")
    create_store(
        "fs",
        {"type": "fs", "path": "this/is/a/relative/fspath", "create_if_missing": True},
    )
    mock_fs.assert_called_once_with("this/is/a/relative/fspath")
    mock_makedirs.assert_called_once_with("this/is/a/relative/fspath")


def test_create_store_mem(mocker):
    mock_mem = mocker.patch("minimalkv.memory.DictStore")
    create_store(
        "memory",
        {"type": "memory", "wrap": "readonly"},
    )
    mock_mem.assert_called_once_with()


def test_create_store_hmem(mocker):
    mock_hmem = mocker.patch("minimalkv._hstores.HDictStore")
    create_store(
        "hmemory",
        {"type": "memory", "wrap": "readonly"},
    )
    mock_hmem.assert_called_once_with()


def test_create_store_redis(mocker):
    mock_Strictredis = mocker.patch("redis.StrictRedis")
    create_store(
        "redis",
        {"type": "redis", "host": "localhost", "db": 2},
    )
    mock_Strictredis.assert_called_once_with(db=2, host="localhost", type="redis")


def test_create_store_valueerror():
    with pytest.raises(Exception, match="Unknown store type: ABC"):
        create_store(
            "ABC",
            {
                "account_name": "ACCOUNT",
                "account_key": "KEY",
                "container": "cont_name",
                "create_if_missing": True,
                "use_sas": True,
            },
        )



import io
from typing import IO, TYPE_CHECKING, Iterator, Optional, Union

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem
    from fsspec.spec import AbstractBufferedFile

from minimalkv import KeyValueStore
from minimalkv.net._net_common import LAZY_PROPERTY_ATTR_PREFIX, lazy_property

# The complete path of the key is structured as follows:
# /Users/simon/data/mykvstore/file1
# <prefix>                    <key>
# If desired to be a directory, the prefix should end in a slash.


class FSSpecStoreEntry(io.BufferedIOBase):
    """A file-like object for reading from an entry in an FSSpecStore."""

    def __init__(self, file: "AbstractBufferedFile"):
        """
        Initialize an FSSpecStoreEntry.

        Parameters
        ----------
        file: AbstractBufferedFile
            The fsspec file object to wrap.
        """
        super().__init__()
        self._file = file

    def seek(self, loc: int, whence: int = 0) -> int:
        """
        Set current file location.

        Parameters
        ----------
        loc: int
            byte location
        whence: {0, 1, 2}
            from start of file, current location or end of file, respectively.
        """
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        try:
            return self._file.seek(loc, whence)
        except ValueError:
            # Map ValueError to IOError
            raise OSError

    def tell(self) -> int:
        """Return the current offset as int. Always >= 0."""
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        return self._file.tell()

    def read(self, size: Optional[int] = -1) -> bytes:
        """Return first ``size`` bytes of data.

        If no size is given all data is returned.

        Parameters
        ----------
        size : int, optional, default = -1
            Number of bytes to be returned.

        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return self._file.read(size)

    def seekable(self) -> bool:
        """Whether the file is seekable."""
        return self._file.seekable()

    def readable(self) -> bool:
        """Whether the file is readable."""
        return self._file.readable()


class FSSpecStore(KeyValueStore):
    """A KeyValueStore that uses an fsspec AbstractFileSystem to store the key-value pairs."""

    def __init__(self, prefix: str = "", mkdir_prefix: bool = True):
        """
        Initialize an FSSpecStore.

        The underlying fsspec FileSystem is created when the store is used for the first time.

        Parameters
        ----------
        prefix: str, optional
            The prefix to use on the FSSpecStore when storing keys.
        mkdir_prefix : Boolean
            If True, the prefix will be created if it does not exist.
            Analogous to the create_if_missing parameter in AzureBlockBlobStore or GoogleCloudStore.
        """
        self.prefix = prefix
        self.mkdir_prefix = mkdir_prefix

    @lazy_property
    def _prefix_exists(self) -> Union[None, bool]:
        from google.auth.exceptions import RefreshError

        # Check if prefix exists.
        # Used by inheriting classes to check if e.g. a bucket exists.
        try:
            return self._fs.exists(self.prefix)
        except (OSError, RefreshError):
            return None

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix: str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        # List files
        all_files_and_dirs = self._fs.find(f"{self.prefix}", prefix=prefix)

        return map(
            lambda k: k.replace(f"{self.prefix}", ""),
            all_files_and_dirs,
        )

    def _delete(self, key: str) -> None:
        try:
            self._fs.rm_file(f"{self.prefix}{key}")
        except FileNotFoundError:
            pass

    def _open(self, key: str) -> IO:
        try:
            return self._fs.open(f"{self.prefix}{key}")
        except FileNotFoundError:
            raise KeyError(key)

    # Required to prevent error when credentials are not sufficient for listing objects
    def _get_file(self, key: str, file: IO) -> str:
        try:
            file.write(self._fs.cat_file(f"{self.prefix}{key}"))
            return key
        except FileNotFoundError:
            raise KeyError(key)

    def _put_file(self, key: str, file: IO) -> str:
        self._fs.pipe_file(f"{self.prefix}{key}", file.read())
        return key

    def _has_key(self, key: str) -> bool:
        return self._fs.exists(f"{self.prefix}{key}")

    def _create_filesystem(self) -> "AbstractFileSystem":
        # To be implemented by inheriting classes.
        raise NotImplementedError

    @lazy_property
    def _fs(self) -> "AbstractFileSystem":
        fs = self._create_filesystem()

        if self.mkdir_prefix and not fs.exists(self.prefix):
            fs.mkdir(self.prefix)
        return fs

    # Skips lazy properties.
    # These will be recreated after unpickling through the lazy_property decorator
    def __getstate__(self) -> dict:  # noqa D
        return {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith(LAZY_PROPERTY_ATTR_PREFIX)
        }



from contextlib import contextmanager
from typing import IO, Dict, Iterator, cast

from minimalkv import CopyMixin, KeyValueStore, UrlMixin


@contextmanager
def map_boto_exceptions(key=None, exc_pass=()):
    """Map boto-specific exceptions to the minimalkv-API."""
    from boto.exception import BotoClientError, BotoServerError, StorageResponseError

    try:
        yield
    except StorageResponseError as e:
        if e.code == "NoSuchKey":
            raise KeyError(key)
        raise OSError(str(e))
    except (BotoClientError, BotoServerError) as e:
        if e.__class__.__name__ not in exc_pass:
            raise OSError(str(e))


class BotoStore(KeyValueStore, UrlMixin, CopyMixin):  # noqa D
    def __init__(
        self,
        bucket,
        prefix="",
        url_valid_time=0,
        reduced_redundancy=False,
        public=False,
        metadata=None,
    ):
        self.prefix = prefix.strip().lstrip("/")
        self.bucket = bucket
        self.reduced_redundancy = reduced_redundancy
        self.public = public
        self.url_valid_time = url_valid_time
        self.metadata = metadata or {}

    def __new_key(self, name):
        from boto.s3.key import Key

        k = Key(self.bucket, self.prefix + name)
        if self.metadata:
            k.update_metadata(self.metadata)
        return k

    def __upload_args(self) -> Dict[str, str]:
        """
        Generate a dictionary of arguments to pass to ``set_content_from`` functions.

        This allows us to save API calls by passing the necessary parameters on with the
        upload.
        """
        d = {
            "reduced_redundancy": self.reduced_redundancy,
        }

        if self.public:
            d["policy"] = "public-read"

        return d

    def iter_keys(self, prefix="") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        with map_boto_exceptions():
            prefix_len = len(self.prefix)
            return map(
                lambda k: k.name[prefix_len:], self.bucket.list(self.prefix + prefix)
            )

    def _has_key(self, key: str) -> bool:
        with map_boto_exceptions(key=key):
            return bool(self.bucket.get_key(self.prefix + key))

    def _delete(self, key: str) -> None:
        from boto.exception import StorageResponseError

        try:
            self.bucket.delete_key(self.prefix + key)
        except StorageResponseError as e:
            if e.code != "NoSuchKey":
                raise OSError(str(e))

    def _get(self, key: str) -> bytes:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            return k.get_contents_as_string()

    def _get_file(self, key: str, file: IO) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            return k.get_contents_to_file(file)

    def _get_filename(self, key: str, filename: str) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            return k.get_contents_to_filename(filename)

    def _open(self, key: str) -> IO:
        from boto.s3.keyfile import KeyFile

        class SimpleKeyFile(KeyFile):  # noqa D
            def read(self, size: int = -1):  # noqa D
                if self.closed:
                    raise ValueError("I/O operation on closed file")
                if size < 0:
                    size = self.key.size - self.location
                return KeyFile.read(self, size)

            def seekable(self):  # noqa D
                return False

            def readable(self):  # noqa D
                return True

        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            return cast(IO, SimpleKeyFile(k))

    def _copy(self, source: str, dest: str) -> None:
        if not self._has_key(source):
            raise KeyError(source)
        with map_boto_exceptions(key=source):
            self.bucket.copy_key(
                self.prefix + dest, self.bucket.name, self.prefix + source
            )

    def _put(self, key: str, data: bytes) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            k.set_contents_from_string(data, **self.__upload_args())
            return key

    def _put_file(self, key: str, file: IO) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            k.set_contents_from_file(file, **self.__upload_args())
            return key

    def _put_filename(self, key: str, filename: str) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            k.set_contents_from_filename(filename, **self.__upload_args())
            return key

    def _url_for(self, key: str) -> str:
        k = self.__new_key(key)
        with map_boto_exceptions(key=key):
            return k.generate_url(expires_in=self.url_valid_time, query_auth=False)



#!/usr/bin/env python

import hashlib
import hmac
import os
import tempfile

from minimalkv.decorator import StoreDecorator


class _HMACFileReader:
    def __init__(self, hm, source):
        self.hm = hm
        self.source = source

        # "preload" buffer
        self.buffer = source.read(self.hm.digest_size)
        if not len(self.buffer) == self.hm.digest_size:
            raise VerificationException(
                "Source does not contain HMAC hash " "(too small)"
            )

    def read(self, n=None):
        if b"" == self.buffer or 0 == n:
            return b""

        new_read = self.source.read(n) if n is not None else self.source.read()
        finished = n is None or len(new_read) != n
        self.buffer += new_read

        if n is not None:
            offset = min(n, len(self.buffer) - self.hm.digest_size)
        else:
            offset = len(self.buffer) - self.hm.digest_size

        rv, self.buffer = self.buffer[:offset], self.buffer[offset:]

        # update hmac
        self.hm.update(rv)

        if finished:
            # check hash
            if not self.buffer == self.hm.digest():
                raise VerificationException("HMAC verification failed.")

        return rv

    def close(self):
        self.source.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class VerificationException(Exception):
    """Exception thrown if there was an error checking authenticity."""

    pass


class HMACDecorator(StoreDecorator):
    """HMAC authentication and integrity check decorator.

    This decorator overrides the :meth:`.KeyValueStore.get`,
    :meth:`.KeyValueStore.get_file`, :meth:`.KeyValueStore.open`,
    :meth:`.KeyValueStore.put` and :meth:`.KeyValueStore.put_file` methods and
    alters the data that is store in the follow way:

    First, the original data is stored while being fed to an hmac instance. The
    resulting hash is appended to the data as a binary string, every value
    stored therefore takes up an additional ``hmac_digestsize`` bytes.

    Upon retrieval using any of :meth:`.KeyValueStore.get`,
    :meth:`.KeyValueStore.get_file` or :meth:`.KeyValueStore.open` methods, the
    data is checked as soon as the hash is readable. Since hashes are stored at
    the end, almost no extra memory is used when using streaming methods.
    However, :meth:`.KeyValueStore.get_file` and :meth:`.KeyValueStore.open`
    will only check the hash value once it is read, that is, at the end of the
    retrieval.

    The decorator will protect against any modification of the stored data and
    ensures that only those with knowledge of the ``__secret_key``
    can alter any data. The key used to store data is also used to extend the
    HMAC secret key, making it impossible to copy a valid message over to a
    different key.
    """

    def __init__(self, secret_key, decorated_store, hashfunc=hashlib.sha256):
        super().__init__(decorated_store)

        self.__hashfunc = hashfunc
        self.__secret_key = bytes(secret_key)

    def __new_hmac(self, key, msg=None):
        # TODO: Comment / Docstring to describe function
        if not msg:
            msg = b""

        # item key is used as salt for secret_key
        hm = hmac.HMAC(
            key=key.encode("ascii") + self.__secret_key,
            msg=msg,
            digestmod=self.__hashfunc,
        )

        return hm

    def get(self, key):  # noqa D
        buf = self._dstore.get(key)
        hm = self.__new_hmac(key)
        hash = buf[-hm.digest_size :]

        # shorten buf
        buf = buf[: -hm.digest_size]

        hm.update(buf)

        if not hm.digest() == hash:
            raise VerificationException("Invalid hash on key %r" % key)

        return buf

    def get_file(self, key, file):  # noqa D
        if isinstance(file, str):
            try:
                f = open(file, "wb")
            except OSError as e:
                raise OSError(f"Error opening {file} for writing: {e!r}")

            # file is open, now we call ourself again with a proper file
            try:
                self.get_file(key, f)
            finally:
                f.close()
        else:
            # need to use open, no way around it it seems
            # this will check the HMAC as well
            source = self.open(key)

            bufsize = 1024 * 1024

            # copy
            while True:
                buf = source.read(bufsize)
                file.write(buf)

                if len(buf) != bufsize:
                    break

    def open(self, key):  # noqa D
        source = self._dstore.open(key)
        return _HMACFileReader(self.__new_hmac(key), source)

    def put(self, key, value, *args, **kwargs):  # noqa D
        # just append hmac and put
        data = value + self.__new_hmac(key, value).digest()
        return self._dstore.put(key, data, *args, **kwargs)  # type: ignore

    def copy(self, source, dest):  # noqa D
        raise NotImplementedError

    def put_file(self, key, file, *args, **kwargs):  # noqa D
        hm = self.__new_hmac(key)
        bufsize = 1024 * 1024

        if isinstance(file, str):
            # we read the file once, then write the hash at the end, before
            # handing it over to the original backend
            with open(file, "rb+") as source:
                while True:
                    buf = source.read(bufsize)
                    hm.update(buf)

                    if len(buf) < bufsize:
                        break

                # file has been read, append hash
                source.write(hm.digest())

            # after the file has been closed, hand it over
            return self._dstore.put_file(key, file, *args, **kwargs)  # type: ignore
        else:
            tmpfile = tempfile.NamedTemporaryFile(delete=False)
            try:
                while True:
                    buf = file.read(bufsize)
                    hm.update(buf)
                    tmpfile.write(buf)

                    if len(buf) < bufsize:
                        break

                tmpfile.write(hm.digest())
                tmpfile.close()

                return self._dstore.put_file(key, tmpfile.name, *args, **kwargs)  # type: ignore
            finally:
                os.unlink(tmpfile.name)



#!/usr/bin/env python

import pytest

from minimalkv.decorator import ReadOnlyDecorator
from minimalkv.memory import DictStore


class TestReadOnlyDecorator:
    def test_readonly(self):
        store0 = DictStore()
        store0.put("file1", b"content")

        store = ReadOnlyDecorator(store0)
        with pytest.raises(AttributeError):
            store.put("file1", b"content2")
        with pytest.raises(AttributeError):
            store.delete("file1")
        assert store.get("file1") == b"content"
        assert "file1" in store
        assert set(store.keys()) == {"file1"}



"""Implement the AzureBlockBlobStore for ``azure-storage-blob<12``."""
import io
from contextlib import contextmanager

from minimalkv._key_value_store import KeyValueStore
from minimalkv.net._azurestore_common import _byte_buffer_md5, _file_md5, _filename_md5
from minimalkv.net._net_common import LAZY_PROPERTY_ATTR_PREFIX, lazy_property


@contextmanager
def map_azure_exceptions(key=None, exc_pass=()):
    """Map Azure-specific exceptions to the minimalkv-API."""
    from azure.common import (
        AzureException,
        AzureHttpError,
        AzureMissingResourceHttpError,
    )

    try:
        yield
    except AzureMissingResourceHttpError as ex:
        if ex.__class__.__name__ not in exc_pass:
            s = str(ex)
            if s.startswith("The specified container does not exist."):
                raise OSError(s)
            raise KeyError(key)
    except AzureHttpError as ex:
        if ex.__class__.__name__ not in exc_pass:
            raise OSError(str(ex))
    except AzureException as ex:
        if ex.__class__.__name__ not in exc_pass:
            raise OSError(str(ex))


class AzureBlockBlobStore(KeyValueStore):  # noqa D
    def __init__(
        self,
        conn_string=None,
        container=None,
        public=False,
        create_if_missing=True,
        max_connections=2,
        max_block_size=None,
        max_single_put_size=None,
        checksum=False,
        socket_timeout=None,
    ):
        self.conn_string = conn_string
        self.container = container
        self.public = public
        self.create_if_missing = create_if_missing
        self.max_connections = max_connections
        self.max_single_put_size = max_single_put_size
        self.max_block_size = max_block_size
        self.checksum = checksum
        self.socket_timeout = socket_timeout

    # Using @lazy_property will (re-)create block_blob_service instance needed.
    # Together with the __getstate__ implementation below, this allows
    # AzureBlockBlobStore to be pickled, even if
    # azure.storage.blob.BlockBlobService does not support pickling.
    @lazy_property
    def block_blob_service(self):  # noqa D
        from azure.storage.blob import BlockBlobService, PublicAccess

        block_blob_service = BlockBlobService(
            connection_string=self.conn_string,
            socket_timeout=self.socket_timeout,
        )
        if self.max_block_size is not None:
            block_blob_service.MAX_BLOCK_SIZE = self.max_block_size
        if self.max_block_size is not None:
            block_blob_service.MAX_SINGLE_PUT_SIZE = self.max_single_put_size

        if self.create_if_missing:
            block_blob_service.create_container(
                self.container,
                public_access=PublicAccess.Container if self.public else None,
            )
        return block_blob_service

    def _delete(self, key: str) -> None:
        with map_azure_exceptions(key=key, exc_pass=["AzureMissingResourceHttpError"]):
            self.block_blob_service.delete_blob(self.container, key)

    def _get(self, key):
        with map_azure_exceptions(key=key):
            return self.block_blob_service.get_blob_to_bytes(
                container_name=self.container,
                blob_name=key,
                max_connections=self.max_connections,
            ).content

    def _has_key(self, key):
        with map_azure_exceptions(key=key):
            return self.block_blob_service.exists(self.container, key)

    def iter_keys(self, prefix=""):  # noqa D
        if prefix == "":
            prefix = None
        with map_azure_exceptions():
            blobs = self.block_blob_service.list_blob_names(
                self.container, prefix=prefix
            )
            return (
                blob.decode("utf-8") if isinstance(blob, bytes) else blob
                for blob in blobs
            )

    def iter_prefixes(self, delimiter, prefix=""):  # noqa D
        if prefix == "":
            prefix = None
        with map_azure_exceptions():
            blobs = self.block_blob_service.list_blob_names(
                self.container, prefix=prefix, delimiter=delimiter
            )
            return (
                blob.decode("utf-8") if isinstance(blob, bytes) else blob
                for blob in blobs
            )

    def _open(self, key):
        with map_azure_exceptions(key=key):
            return IOInterface(
                self.block_blob_service, self.container, key, self.max_connections
            )

    def _put(self, key, data):
        from azure.storage.blob.models import ContentSettings

        if self.checksum:
            content_settings = ContentSettings(content_md5=_byte_buffer_md5(data))
        else:
            content_settings = ContentSettings()

        with map_azure_exceptions(key=key):
            self.block_blob_service.create_blob_from_bytes(
                container_name=self.container,
                blob_name=key,
                blob=data,
                max_connections=self.max_connections,
                content_settings=content_settings,
            )
            return key

    def _put_file(self, key, file):
        from azure.storage.blob.models import ContentSettings

        if self.checksum:
            content_settings = ContentSettings(content_md5=_file_md5(file))
        else:
            content_settings = ContentSettings()

        with map_azure_exceptions(key=key):
            self.block_blob_service.create_blob_from_stream(
                container_name=self.container,
                blob_name=key,
                stream=file,
                max_connections=self.max_connections,
                content_settings=content_settings,
            )
            return key

    def _get_file(self, key, file):
        with map_azure_exceptions(key=key):
            self.block_blob_service.get_blob_to_stream(
                container_name=self.container,
                blob_name=key,
                stream=file,
                max_connections=self.max_connections,
            )

    def _get_filename(self, key, filename):
        with map_azure_exceptions(key=key):
            self.block_blob_service.get_blob_to_path(
                container_name=self.container,
                blob_name=key,
                file_path=filename,
                max_connections=self.max_connections,
            )

    def _put_filename(self, key, filename):
        from azure.storage.blob.models import ContentSettings

        if self.checksum:
            content_settings = ContentSettings(content_md5=_filename_md5(filename))
        else:
            content_settings = ContentSettings()

        with map_azure_exceptions(key=key):
            self.block_blob_service.create_blob_from_path(
                container_name=self.container,
                blob_name=key,
                file_path=filename,
                max_connections=self.max_connections,
                content_settings=content_settings,
            )
            return key

    def __getstate__(self):  # noqa D
        # keep all of __dict__, except lazy properties:
        return {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith(LAZY_PROPERTY_ATTR_PREFIX)
        }


class IOInterface(io.BufferedIOBase):
    """File-like interface to selectively read from a blob in the blob store."""

    def __init__(self, block_blob_service, container_name, key, max_connections):
        super().__init__()
        self.block_blob_service = block_blob_service
        self.container_name = container_name
        self.key = key
        self.max_connections = max_connections

        blob = self.block_blob_service.get_blob_properties(container_name, key)
        self.size = blob.properties.content_length
        self.pos = 0

    def tell(self):
        """Return the current offset as int. Always >= 0."""
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return self.pos

    def read(self, size=-1):
        """Return 'size' amount of bytes or less if there is no more data.

        If no size is given all data is returned. size can be >= 0.
        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        with map_azure_exceptions(key=self.key):
            if size < 0:
                size = self.size - self.pos

            end = min(self.pos + size - 1, self.size - 1)
            if self.pos > end:
                return b""
            b = self.block_blob_service.get_blob_to_bytes(
                container_name=self.container_name,
                blob_name=self.key,
                start_range=self.pos,
                end_range=end,  # end_range is inclusive
                max_connections=self.max_connections,
            )
            self.pos += len(b.content)
            return b.content

    def seek(self, offset, whence=0):
        """Move to a new offset either relative or absolute.

        whence=0 is absolute, whence=1 is relative, whence=2 is relative to the
        end.

        Any relative or absolute seek operation which would result in a
        negative position is undefined and that case can be ignored
        in the implementation.

        Any seek operation which moves the position after the stream
        should succeed. tell() should report that position and read()
        should return an empty bytes object.

        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        if whence == 0:
            if offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos = offset
        elif whence == 1:
            if self.pos + offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos += offset
        elif whence == 2:
            if self.size + offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos = self.size + offset
        return self.pos

    def seekable(self):  # noqa
        return True

    def readable(self):  # noqa
        return True



import os

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.fs import FilesystemStore
from minimalkv.memory import DictStore
from minimalkv.memory.redisstore import RedisStore
from minimalkv.net.azurestore import AzureBlockBlobStore
from minimalkv.net.botostore import BotoStore
from minimalkv.net.gcstore import GoogleCloudStore


class HDictStore(ExtendedKeyspaceMixin, DictStore):  # noqa D
    pass


class HRedisStore(ExtendedKeyspaceMixin, RedisStore):  # noqa D
    pass


class HAzureBlockBlobStore(ExtendedKeyspaceMixin, AzureBlockBlobStore):  # noqa D
    pass


class HBotoStore(ExtendedKeyspaceMixin, BotoStore):  # noqa D
    def size(self, key: str) -> bytes:
        """Get size of data at key in bytes.

        Parameters
        ----------
        key : str
            Key of data.

        Returns
        -------
        size : int
            Size of value at key in bytes.
        """
        k = self.bucket.lookup(self.prefix + key)
        return k.size


class HGoogleCloudStore(ExtendedKeyspaceMixin, GoogleCloudStore):  # noqa D
    pass


class HFilesystemStore(ExtendedKeyspaceMixin, FilesystemStore):  # noqa D
    def size(self, key: str) -> int:
        """Get size of data at key in bytes.

        Parameters
        ----------
        key : str
            Key of data.

        Returns
        -------
        size : int
            Size of value at key in bytes.
        """
        return os.path.getsize(self._build_filename(key))



import json
import warnings
from typing import IO, cast

from minimalkv.fsspecstore import FSSpecStore, FSSpecStoreEntry

try:
    from gcsfs import GCSFileSystem

    has_gcsfs = True
except ImportError:
    has_gcsfs = False


class GoogleCloudStore(FSSpecStore):
    """A store using ``Google Cloud storage`` as a backend.

    See ``https://cloud.google.com/storage``.
    """

    def __init__(
        self,
        credentials,
        bucket_name: str,
        create_if_missing: bool = True,
        bucket_creation_location: str = "EUROPE-WEST3",
        project=None,
    ):
        if isinstance(credentials, str):
            # Parse JSON from path to extract project name
            # The project name is required to create new buckets
            try:
                with open(credentials) as f:
                    credentials_dict = json.load(f)
                    project = project or credentials_dict["project_id"]
            except (FileNotFoundError, json.JSONDecodeError, KeyError) as error:
                warnings.warn(
                    f"""
                    Could not get the project name from the credentials file.
                    You set create_if_missing to {create_if_missing}.
                    You will not be able to create a new bucket for this store.

                    This was caused by the following error:
                    {error}
                    """
                )

        self._credentials = credentials
        self.bucket_name = bucket_name
        self.create_if_missing = create_if_missing
        self.bucket_creation_location = bucket_creation_location
        self.project_name = project

        super().__init__(prefix=f"{bucket_name}/", mkdir_prefix=create_if_missing)

    def _create_filesystem(self) -> "GCSFileSystem":
        if not has_gcsfs:
            raise ImportError("Cannot find optional dependency gcsfs.")

        return GCSFileSystem(
            project=self.project_name,
            token=self._credentials,
            access="read_write",
            default_location=self.bucket_creation_location,
        )

    def _open(self, key: str) -> IO:
        from google.cloud.exceptions import NotFound

        if self._prefix_exists is False:
            raise NotFound(f"Could not find bucket: {self.bucket_name}")
        return cast(IO, FSSpecStoreEntry(super()._open(key)))

    def _get_file(self, key: str, file: IO) -> str:
        from google.cloud.exceptions import NotFound

        if self._prefix_exists is False:
            raise NotFound(f"Could not find bucket: {self.bucket_name}")
        return super()._get_file(key, file)



#!/usr/bin/env python

from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="minimalkv",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description=("A key-value storage for binary data, support many " "backends."),
    long_description=long_description,
    author="Data Engineering Collective",
    author_email="minimalkv@uwekorn.com",
    url="https://github.com/data-engineering-collective/minimalkv",
    license="BSD-3-clause",
    packages=find_packages(exclude=["test"]),
    install_requires=["uritools"],
    python_requires=">=3.8",
    package_data={"minimalkv": ["py.typed"]},
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
    ],
)



def _get_s3bucket(
    host,
    bucket,
    access_key,
    secret_key,
    force_bucket_suffix=True,
    create_if_missing=True,
):
    # TODO: Write docstring.
    from boto.s3.connection import S3ResponseError  # type: ignore
    from boto.s3.connection import OrdinaryCallingFormat, S3Connection

    s3con = S3Connection(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        host=host,
        is_secure=False,
        calling_format=OrdinaryCallingFormat(),
    )
    # add access key prefix to bucket name, unless explicitly prohibited
    if force_bucket_suffix and not bucket.lower().endswith("-" + access_key.lower()):
        bucket = bucket + "-" + access_key.lower()
    try:
        return s3con.get_bucket(bucket)
    except S3ResponseError as ex:
        if ex.status == 404:
            if create_if_missing:
                return s3con.create_bucket(bucket)
            else:
                raise OSError(f"Bucket {bucket} does not exist")
        raise



from io import BytesIO
from types import TracebackType
from typing import IO, Iterable, Iterator, List, Optional, Type, Union

from minimalkv._constants import VALID_KEY_RE
from minimalkv._mixins import UrlMixin

# Only here to keep backwards-compatability
key_type = str


class KeyValueStore:
    """
    Class to access a key-value store.

    Supported keys are ascii-strings containing alphanumeric characters or symbols out
    of ``minimalkv._constants.VALID_NON_NUM`` of length not greater than 250. Values
    (or records) are stored as raw bytes.
    """

    def __contains__(self, key: str) -> bool:
        """Check if the store has an entry at key.

        Parameters
        ----------
        key : str
            The key whose existence should be verified.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was an error accessing the store.
        """
        self._check_valid_key(key)
        return self._has_key(key)

    def __iter__(self) -> Iterable[str]:
        """Iterate over all keys in the store.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        return self.iter_keys()

    def delete(self, key: str) -> Optional[str]:
        """Delete data at key.

        Does not raise an error if the key does not exist.

        Parameters
        ----------
        key: str
            The key of data to be deleted.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was an error deleting.
        """
        self._check_valid_key(key)
        return self._delete(key)

    def get(self, key: str) -> bytes:
        """Return data at key as a bytestring.

        Parameters
        ----------
        key : str
            The key to be read.

        Returns
        -------
        data : str
            Value associated with the key as a ``bytes`` object.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If the file could not be read.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        return self._get(key)

    def get_file(self, key: str, file: Union[str, IO]) -> str:
        """Write data at key to file.

        Like :meth:`~mininmalkv.KeyValueStore.put_file`, this method allows backends to
        implement a specialized function if data needs to be written to disk or streamed.

        If ``file`` is a string, contents of ``key`` are written to a newly created file
        with the filename ``file``. Otherwise the data will be written using the
        ``write`` method of ``file``.

        Parameters
        ----------
        key : str
            The key to be read.
        file : file-like or str
            Output filename or file-like object with a ``write`` method.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem reading or writing data.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        if isinstance(file, str):
            return self._get_filename(key, file)
        else:
            return self._get_file(key, file)

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        raise NotImplementedError

    def iter_prefixes(self, delimiter: str, prefix: str = "") -> Iterator[str]:
        """
        Iterate over unique prefixes in the store up to delimiter, starting with prefix.

        If ``prefix`` contains ``delimiter``, return the prefix up to the first
        occurence of delimiter after the prefix.

        The default uses an naive key iteration. Some backends may implement more
        efficient methods.

        Parameters
        ----------
        delimiter : str, optional, default = ''
            Delimiter up to which to iterate over prefixes.
        prefix : str, optional, default = ''
            Only iterate over prefixes starting with prefix.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        dlen = len(delimiter)
        plen = len(prefix)
        memory = set()

        for k in self.iter_keys(prefix):
            pos = k.find(delimiter, plen)
            if pos >= 0:
                k = k[: pos + dlen]

            if k not in memory:
                yield k
                memory.add(k)

    def keys(self, prefix: str = "") -> List[str]:
        """List all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only list keys starting with prefix. List all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        return list(self.iter_keys(prefix))

    def open(self, key: str) -> IO:
        """Open record at key.

        Parameters
        ----------
        key : str
            Key to open.

        Returns
        -------
        file: file-like
            Read-only file-like object for reading data at key.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If the file could not be read.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        return self._open(key)

    def put(self, key: str, data: bytes) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str
            The key under which the data is to be stored.
        data : bytes
            Data to be stored at key, must be of type  ``bytes``.

        Returns
        -------
        str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If storing failed or the file could not be read.
        """
        self._check_valid_key(key)
        if not isinstance(data, bytes):
            raise OSError("Provided data is not of type bytes")
        return self._put(key, data)

    def put_file(self, key: str, file: Union[str, IO]) -> str:
        """Store contents of file at key.

        Store data from a file into key. ``file`` can be a string, which will be
        interpreted as a filename, or an object with a ``read()`` method.

        If ``file`` is a filename, the file might be removed while storing to avoid
        unnecessary copies. To prevent this, pass the opened file instead.

        Parameters
        ----------
        key : str
            Key where to store data in file.
        file : file-like or str
            A filename or a file-like object with a read method.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem moving the file in.

        """
        self._check_valid_key(key)
        if isinstance(file, str):
            return self._put_filename(key, file)
        else:
            return self._put_file(key, file)

    def _check_valid_key(self, key: str) -> None:
        """Check if a key is valid and raise a ValueError if it is not.

        Always use this method to check whether a key is valid.

        Parameters
        ----------
        key : str
            The key to be checked.

        Raises
        ------
        ValueError
            If the key is not valid.
        """
        if not isinstance(key, key_type):
            raise ValueError(f"The key {key} is not a valid key type.")
        if not VALID_KEY_RE.match(key):
            raise ValueError(f"The key {key} contains illegal characters.")

    def _delete(self, key: str):
        """Delete the data at key in store."""
        raise NotImplementedError

    def _get(self, key: str) -> bytes:
        """Read data at key in store.

        Parameters
        ----------
        key : str
            Key of value to be retrieved.
        """
        buf = BytesIO()

        self._get_file(key, buf)

        return buf.getvalue()

    def _get_file(self, key: str, file: IO) -> str:
        """Write data at key to file-like object file.

        Parameters
        ----------
        key : str
            Key of data to be written to file.
        file : file-like
            File-like object with a *write* method to be written.
        """
        bufsize = 1024 * 1024

        # note: we do not use a context manager here or close the source.
        # the source goes out of scope shortly after, taking care of the issue
        # this allows us to support file-like objects without close as well,
        # such as BytesIO.
        source = self.open(key)
        try:
            while True:
                buf = source.read(bufsize)
                file.write(buf)

                if len(buf) < bufsize:
                    break
        finally:
            source.close()

        return key

    def _get_filename(self, key: str, filename: str) -> str:
        """Write data at key to file at filename.

        Parameters
        ----------
        key : str
            Key of data to be written to file at filename.
        filename : str
            Name of file to be written.
        """
        with open(filename, "wb") as dest:
            return self._get_file(key, dest)

    def _has_key(self, key: str) -> bool:
        """Check the existence of key in store.

        Parameters
        ----------
        key : str
            Key to check the existance of.
        """
        return key in self.keys()

    def _open(self, key: str) -> IO:
        """Open record at key.

        Parameters
        ----------
        key : str
            Key of record to open.

        Returns
        -------
        file: file-like
            Opened file.
        """
        raise NotImplementedError

    def _put(self, key: str, data: bytes) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        data : bytes
            Data to be stored.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        return self._put_file(key, BytesIO(data))

    def _put_file(self, key: str, file: IO) -> str:
        """Store data from file-like object at key.

        Parameters
        ----------
        key : str
            Key at which to store contents of file.
        file : file-like
            File-like object to store data from.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        raise NotImplementedError

    def _put_filename(self, key: str, filename: str) -> str:
        """Store data from file at filename at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        filename : str
            Filename of file to store.

        Returns
        -------
        key: str
            Key where data was stored.

        """
        with open(filename, "rb") as source:
            return self._put_file(key, source)

    def close(self):
        """Clean up all open resources in child classes.

        Specific store implementations might require teardown methods
        (dangling ports, unclosed files). This allows calling close also
        for stores, which do not require this.
        """
        return

    def __enter__(self):
        """Support with clause for automatic calling of close."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        """Support with clause for automatic calling of close.

        :param exc_type: Type of optional exception encountered in context manager
        :param exc_val: Actual optional exception encountered in context manager
        :param exc_tb: Traceback of optional exception encountered in context manager
        """
        self.close()


class UrlKeyValueStore(UrlMixin, KeyValueStore):
    """Class is deprecated. Use the :class:`.UrlMixin` instead.

    .. deprecated:: 0.9

    """

    pass



#!/usr/bin/env python

import pytest

sqlalchemy = pytest.importorskip("sqlalchemy")
from basic_store import BasicStore
from conftest import ExtendedKeyspaceTests
from sqlalchemy import MetaData, create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.pool import StaticPool

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.db.sql import SQLAlchemyStore

DSNS = [
    (
        "pymysql",
        "mysql+pymysql://minimalkv_test:minimalkv_test@127.0.0.1/minimalkv_test",
    ),
    (
        "psycopg2",
        "postgresql+psycopg2://minimalkv_test:minimalkv_test@127.0.0.1/minimalkv_test",
    ),
    ("sqlite3", "sqlite:///:memory:"),
]


# FIXME: for local testing, this needs configurable dsns
class TestSQLAlchemyStore(BasicStore):
    @pytest.fixture(params=DSNS, ids=[v[0] for v in DSNS])
    def engine(self, request):
        module_name, dsn = request.param
        # check module is available
        pytest.importorskip(module_name)
        engine = create_engine(dsn, poolclass=StaticPool)
        try:
            engine.connect()
        except OperationalError:
            pytest.skip(f"could not connect to database {dsn}")
        return engine

    @pytest.fixture
    def store(self, engine):
        metadata = MetaData(bind=engine)
        with SQLAlchemyStore(engine, metadata, "minimalkv_test") as store:
            # create table
            store.table.create()
            yield store
        metadata.drop_all()


class TestExtendedKeyspaceSQLAlchemyStore(TestSQLAlchemyStore, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, engine):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, SQLAlchemyStore):
            pass

        metadata = MetaData(bind=engine)
        with ExtendedKeyspaceStore(engine, metadata, "minimalkv_test") as store:
            # create table
            store.table.create()
            yield store
        metadata.drop_all()
